##########################################################################
# NSAp - Copyright (C) CEA, 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that contains export utilities.
"""

# System import
import os
import json
import gzip


def gzip_and_update_outputs(results):
    """ Compress file and update the result outputs accordingly. Note that
    a simplified results structure as produced by the
    'export_interface_results' function is expected.

    Note that the results structure will be updated (ie modified) by the
    function.

    Parameters
    ----------
    results: dict (mandatory)
        a results structure with process name as keys and process results
        as values. Each process values is expected to contain an 'outputs'
        key containing a dictionary mapping of the process outputs.
    """
    for process_name, results in results.items():
        updated_params = []
        if isinstance(results, (list, tuple)):
            for cnt, item in enumerate(results):
               results[cnt] = recursive_gzip(item["outputs"])
            continue
        for param_name, param in results["outputs"].items():
            results["outputs"][param_name] = recursive_gzip(param)


def recursive_gzip(obj):
    """ Recursively find and Gzip files.

    Parameters
    ----------
    obj: object (mandatory)
        a Python object containing files to Gzip.

    Returns
    -------
    gzip_obj: object
        the input object with Gziped files.
    """
    # Stop case: a non iterative structure is reached
    if isinstance(obj, basestring) and os.path.isfile(obj):
        return gzip_file(obj, prefix="", outdir=None,
                         remove_original_file=True)

    # Go deeper
    elif isinstance(obj, (tuple, list)):
        gzip_obj = []
        for item in obj:
            gzip_obj.append(recursive_gzip(obj))
        if isinstance(obj, tuple):
            gzip_obj = tuple(gzip_obj)
        return gzip_obj

    # Default return object
    else:
        return obj


def ungzip_file(fname, prefix="u", outdir=None):
    """ Copy and ungzip the input file.

    Parameters
    ----------
    fname: str (mandatory)
        an input file to ungzip.
    prefix: str (optional, default 'u')
        the prefix of the result file.
    outdir: str (optional, default None)
        the output directory where ungzip file is saved.

    Returns
    -------
    ungzipfname: str
        the returned ungzip file.
    </unit>
    """
    # Check the input file exists on the file system
    if not os.path.isfile(fname):
        raise ValueError("'{0}' is not a valid filename.".format(fname))

    # Check that the outdir is valid
    if outdir is not None:
        if not os.path.isdir(outdir):
            raise ValueError(
                "'{0}' is not a valid directory.".format(outdir))
    else:
        outdir = os.path.dirname(fname)

    # Get the file descriptors
    base, extension = os.path.splitext(fname)
    basename = os.path.basename(base)

    # Ungzip only known extension
    if extension in [".gz"]:

        # Generate the output file name
        basename = prefix + basename
        ungzipfname = os.path.join(outdir, basename)

        # Read the input gzip file
        with gzip.open(fname, "rb") as gzfobj:
            data = gzfobj.read()

        # Write the output ungzip file
        with open(ungzipfname, "wb") as openfile:
            openfile.write(data)

    # Default, unknown compression extension: the input file is returned
    else:
        ungzipfname = fname

    return ungzipfname


def gzip_file(ungzip_file, prefix="g", outdir=None,
              remove_original_file=False):
    """ Gzip an input file and possibly remove the original file.

    Parameters
    ----------
    ungzip_file: str (mandatory)
        an input file to gzip.
    prefix: str (optional, default 'g')
        a prefix that will be concatenated to the produced file basename.
    outdir: str (optional, default None)
        the destination folder where the Gzip file is saved. If this parameter
        is None, the input image folder is considered as an output folder.
    remove_original_file: bool (optiona, default False)
        if True, remove the original file.

    Returns
    -------
    gzip_file: str
        the returned Gzip file.
    """
    # Check the input file exists on the file system
    if not os.path.isfile(ungzip_file):
        raise ValueError("'{0}' is not a valid filename.".format(ungzip_file))

    # Check that the outdir is valid
    if outdir is not None:
        if not os.path.isdir(outdir):
            raise ValueError("'{0}' is not a valid directory.".format(outdir))
    else:
        outdir = os.path.dirname(ungzip_file)

    # Get the file descriptors
    dirname, basename = os.path.split(ungzip_file)
    base, extension = os.path.splitext(basename)

    # Gzip only non compressed file
    if extension not in [".gz", ".mat", ".json", ".txt"]:

        # Generate the output file name
        basename = base + extension + ".gz"
        if prefix:
            basename = prefix + basename
        gzip_file = os.path.join(outdir, basename)

        # Write the output gzip file
        with open(ungzip_file, "rb") as openfile:
            with gzip.open(gzip_file, "w") as gzfobj:
                gzfobj.writelines(openfile)

        # Remove original file if requested
        if remove_original_file:
            os.remove(ungzip_file)

    # Default, the input file is returned
    else:
        gzip_file = ungzip_file

    return gzip_file




##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################


"""
Module for generating post-preproc plots (registration, segmentation, etc.)
"""


# Imports
import numpy as np
import matplotlib.pyplot as plt


def plot_fsl_motion_parameters(parameter_file, outfname):
    """ Plot motion parameters obtained with FSL software

    Parameters
    ----------
    parameter_file: string
        path of file containing the motion parameters.
    outfname: string
        output filename for storing the plotted figure.

    """
    # Load parameters
    motion = np.loadtxt(parameter_file)
    motion[:, :3] *= (180. / np.pi)

    # do plotting
    plt.figure()
    fig, (ax1, ax2) = plt.subplots(2, 1)
    ax1.plot(motion[:, 3:])
    ax1.set_xlabel("time(scans)")
    ax1.set_ylabel("Estimated motion (mm)")
    ax1.grid(True)
    ax2.plot(motion[:, :3])
    ax2.set_xlabel("time(scans)")
    ax2.set_ylabel("Estimated motion (degrees)")
    ax2.grid(True)
    plt.legend(("TransX", "TransY", "TransZ", "RotX", "RotY", "RotZ"),
               loc="upper left", ncol=2)
    plt.savefig(outfname)
    plt.close()




##########################################################################
# NSAP - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
File utilities.
"""

# Sytem import
import os

# Third party import
from pyconnectome import DEFAULT_FSL_PATH
from pyconnectome.wrapper import FSLWrapper


def average_timeserie(input_file, output_file, fslconfig=DEFAULT_FSL_PATH):
    """ Average a timeserie.

    Parameters
    ----------
    input_file: str
        the image to average.
    output_file: str
        the averaged image.
    fslconfig: str, default DEFAULT_FSL_PATH
        the FSL configuration batch.
    """
    # Check the input parameter
    for path in (input_file, ):
        if not os.path.isfile(path):
            raise ValueError("'{0}' is not a valid input file.".format(path))

    # Define the FSL command
    cmd = ["fslmaths", input_file, "-Tmean", output_file]

    # Call fslmaths
    fslprocess = FSLWrapper(shfile=fslconfig)
    fslprocess(cmd=cmd)




##########################################################################
# NSAp - Copyright (C) CEA, 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that can be used to reorient images.
"""

# System import
import os
import numpy
import nibabel


# Global parameters
POSSIBLE_AXES_ORIENTATIONS = [
    "LAI", "LIA", "ALI", "AIL", "ILA", "IAL",
    "LAS", "LSA", "ALS", "ASL", "SLA", "SAL",
    "LPI", "LIP", "PLI", "PIL", "ILP", "IPL",
    "LPS", "LSP", "PLS", "PSL", "SLP", "SPL",
    "RAI", "RIA", "ARI", "AIR", "IRA", "IAR",
    "RAS", "RSA", "ARS", "ASR", "SRA", "SAR",
    "RPI", "RIP", "PRI", "PIR", "IRP", "IPR",
    "RPS", "RSP", "PRS", "PSR", "SRP", "SPR"
]
CORRECTION_MATRIX_COLUMNS = {
    "R": (1, 0, 0),
    "L": (-1, 0, 0),
    "A": (0, 1, 0),
    "P": (0, -1, 0),
    "S": (0, 0, 1),
    "I": (0, 0, -1)
}


def swap_affine(axes):
    """ Build a correction matrix, from the given orientation of axes to RAS.

    Parameters
    ----------
    axes: str (manadtory)
        the given orientation of the axes.

    Returns
    -------
    rotation: array (4, 4)
        the correction matrix.
    """
    rotation = numpy.eye(4)
    rotation[:3, 0] = CORRECTION_MATRIX_COLUMNS[axes[0]]
    rotation[:3, 1] = CORRECTION_MATRIX_COLUMNS[axes[1]]
    rotation[:3, 2] = CORRECTION_MATRIX_COLUMNS[axes[2]]
    return rotation


def reorient_image(in_file, axes="RAS", prefix="swap", output_directory=None):
    """ Rectify the orientation of an image in order to be in the 'RAS'
    coordinate system.

    Parameters
    ----------
    in_file: str (mandatory)
        the input image.
    axes: str (optional, default 'RAS')
        orientation of the original axes X, Y, and Z
        specified with the following convention: L=Left, R=Right,
        A=Anterion, P=Posterior, I=Inferior, S=Superior.
    prefix: str (optional, default 'swap')
        prefix of the output image.
    output_directory: str (optional, default None)
        the output directory where the rectified image is saved.
        If None use the same directory as the input image.

    Returns
    -------
    out_file: str
        the rectified image.

    Examples
    --------

    >>> from pyfmri.utils.reorient import reorient_image
    >>> rectified_image = reorient_image('image.nii', 'RAS', 's', None)
    """
    # Check the input image exists on the file system
    if not os.path.isfile(in_file):
        raise ValueError("'{0}' is not a valid filename.".format(in_file))

    # Check that the outdir is valid
    if output_directory is not None:
        if not os.path.isdir(output_directory):
            raise ValueError("'{0}' is not a valid directory.".format(
                output_directory))
    else:
        output_directory = os.path.dirname(in_file)

    # Check that a valid input axes is specified
    if axes not in POSSIBLE_AXES_ORIENTATIONS:
        raise ValueError("Wrong '{0}' coordinate system.".format(axes))

    # Get the transformation to the RAS space
    rotation = swap_affine(axes)
    det = numpy.linalg.det(rotation)
    if det != 1:
        print("Rotation matrix determinant must be one not '{0}'.".format(det))

    # Load the image to rectify
    image = nibabel.load(in_file)

    # Get the input image affine transform
    affine = image.get_affine()

    # Apply the rotation to set the image in the RAS coordiante system
    transformation = numpy.dot(rotation, affine)
    image.set_qform(transformation)
    image.set_sform(transformation)

    # Save the rectified image
    basename = os.path.basename(in_file)
    if basename.endswith(".nii"):
        basename += ".gz"
    elif not basename.endswith(".nii.gz"):
        basename += ".nii.gz"
    out_file = os.path.join(output_directory, prefix + basename)
    nibabel.save(image, out_file)

    return out_file


def switch_radiological_neurological(in_file, out_file):
    """ Radiologists like looking at their images with the patient's left
    on the right of the image. If they are looking at a brain image, it is
    as if they were looking at the brain slice from the point of view of
    the patient's feet. Neurologists like looking at brain images with the
    patient's right on the right of the image. This perspective is as if the
    neurologist is looking at the slice from the top of the patient's head.
    The convention is one of image display.

    Parameters
    ----------
    in_file: str (mandatory)
        the input image in RAS convention.
    out_file: str (mandatory)
        the path to the input image with the left/right direction switched.
    """
    im = nibabel.load(in_file)
    switch_im = nibabel.Nifti1Image(
        im.get_data()[::-1, :, :], affine=im.get_affine(),
        header=im.get_header())
    nibabel.save(switch_im, out_file)


def check_orientation(in_files):
    """ Check if the input images have the same orientation.

    Parameters
    ----------
    in_files: str or list of str (mandatory)
        some images.

    Returns
    -------
    test: bool
        true if the orientations are the same.
    """
    if not isinstance(in_files, list):
        in_files = [in_files]
    orients = []
    for path in in_files:
        im = nibabel.load(path)
        orients.append("".join(nibabel.aff2axcodes(im.affine)))
    return all([elem == orients[0] for elem in orients]), orients


def guess_orientation(in_file):
    """ Return an image orientation.

    Parameters
    ----------
    in_files: str or list of str (mandatory)
        an image.

    Returns
    -------
    axes: str
        the image orientation.
    """
    im = nibabel.load(in_file)
    return "".join(nibabel.aff2axcodes(im.affine))
    



#! /usr/bin/env python
##########################################################################
# NSAp - Copyright (C) CEA, 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# pyConnectomist current version
version_major = 1
version_minor = 0
version_micro = 0

# Expected by setup.py: string of form "X.Y.Z"
__version__ = "{0}.{1}.{2}".format(version_major, version_minor, version_micro)

# Define default SPM STANDALONE path for the package
DEFAULT_SPM_STANDALONE_PATH = "/i2bm/local/bin/spm12"

# Define default FSL configuration path for the package
DEFAULT_FSL_PATH = "/etc/fsl/4.1/fsl.sh"

# Expected by setup.py: the status of the project
CLASSIFIERS = ["Development Status :: 5 - Production/Stable",
               "Environment :: Console",
               "Environment :: X11 Applications :: Qt",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering",
               "Topic :: Utilities"]

# Project descriptions
description = """
Python processing for PreClinical data.
"""
SUMMARY = """
.. container:: summary-carousel

    pypreclin is a Python module for **processing preclinical data** (non-human
    primate) that offers:

    * pypreclin_preproc_fmri: a script to pre-process fMRI data.
    * pypreclin_prepare_target: a script to perform B0 ihomogeneities
      correction using a non linear registration to the anatomical image.
    * pypreclin_preproc_multi_fmri: a script to pre-process BIDS fMRI data.
"""
long_description = (
    "pypreclin is a Python project that provides a collection of Python "
    "scripts for processing MRI preclinical datasets. This work is made "
    "available by a community of people, amoung which the CEA Neurospin "
    "UNATI and CEA NeuroSpin UNICOG laboratories, in particular A. Grigis, "
    "J. Tasserie, and B. Jarraya.\n")

# Main setup parameters
NAME = "pypreclin"
ORGANISATION = "CEA"
MAINTAINER = "Antoine Grigis"
MAINTAINER_EMAIL = "antoine.grigis@cea.fr"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "https://github.com/neurospin/pypreclin"
DOWNLOAD_URL = "https://github.com/neurospin/pypreclin"
LICENSE = "CeCILL-B"
CLASSIFIERS = CLASSIFIERS
AUTHOR = """
Antoine Grigis <antoine.grigis@cea.fr>
Jordy Tasserie <jordy.tasserie@cea.fr>
Bechir Jarraya <bechir.jarraya@cea.fr>
"""
AUTHOR_EMAIL = "antoine.grigis@cea.fr"
PLATFORMS = "OS Independent"
ISRELEASE = True
VERSION = __version__
PROVIDES = ["pypreclin"]
REQUIRES = [
    "numpy>=1.11.0",
    "scipy>=0.17.0",
    "hopla>=1.0.5",
    "pyconnectome[standalone]>=1.0.0",
    "pyconnectomist>=2.0.0",
    "nipype>=1.0.1",
    "matplotlib==3.0.3",
    "nibabel>=1.1.0",
    "joblib>=0.13.2",
    "transforms3d>=0.3.1",
    "filelock>=3.0.12"]
EXTRA_REQUIRES = {
    "gui": {
        "python-pypipe>=0.0.1"
    }
}
EXTRANAME = "UNICOG"
EXTRAURL = "http://www.unicog.org/site_2016/"



# -*- coding: utf-8 -*-
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
fMRI preprocessings using FSL, SPM, JIP, and ANTS.

Steps
-----

1. Slice Timing correction (with cache).
2. B0 inhomogeneties correction (optional, with cache).
3. Reorient images not in RAS coordinate system and reorient images to match
   the orientation of the standard MNI152 template (with cache).
4. Realign: motion correction - adjust for movement between slices (with
   cache).
5. Normalization: warp images to fit to a standard template brain (with cache).
6. Tissues segmentation and spatial intensity variations correction (with
   cache).
7. Coregistration: overlay structural and functional images - link
   functional scans to anatomical scan (with cache).
8. Wrap functional: resample the functional serie and mask the registered
   serie (with cache).
9. Smooth the functional time serie (with cache).
10. SNAPs: compute some snaps assessing the different processing steps (with
    cache).
11. Reporting: generate a QC reporting (with cache).
"""

# System import
from __future__ import print_function
import os
import re
import sys
import ast
import json
import glob
import shutil
import tempfile
from pprint import pprint
from datetime import datetime

# Module import
import pypreclin
from pypreclin import DEFAULT_FSL_PATH
from pypreclin.utils.reorient import reorient_image
from pypreclin.utils.reorient import guess_orientation
from pypreclin.preproc.register import check_jip_install
from pypreclin.preproc.register import jip_align
from pypreclin.preproc.register import apply_jip_align
from pypreclin.preproc.register import timeserie_to_reference
from pypreclin.preproc.register import resample_image
from pypreclin.preproc.undist import topup
from pypreclin.preproc.undist import fugue
from pypreclin.plotting.check_preprocessing import plot_fsl_motion_parameters

# PyConnectome import
from pyconnectome.configuration import environment, concat_environment
from pyconnectome.utils.filetools import fslreorient2std
from pyconnectome.utils.filetools import apply_mask
from pyconnectome.utils.segtools import fast
from pyconnectome.utils.segtools import bet2
from pyconnectome.utils.regtools import mcflirt
from pyconnectome.utils.regtools import flirt
from pyconnectome.plotting.slicer import triplanar

# PyConnectomist import
from pyconnectomist.utils.pdftools import generate_pdf

# Third party import
import nibabel
import pylab as plt
from hopla.converter import hopla
from joblib import Memory as JoblibMemory
from nipype.interfaces import fsl
from nipype.interfaces import ants
from nipype.caching import Memory as NipypeMemory


# Global parameters
STEPS = {
    "slice_timing": "1-SliceTiming",
    "warp": "2-B0Inhomogeneities",
    "reorient": "3-Reorient",
    "realign": "4-Realign",
    "normalization": "5-Normalization",
    "inhomogeneities": "6-Inhomogeneities",
    "coregistration": "7-Coregistration",
    "wrap": "8-Wrap",
    "smooth": "9-Smooth",
    "snaps": "10-Snaps",
    "report": "11-Report"
}


def preproc(
        funcfile,
        anatfile,
        sid,
        outdir,
        repetitiontime,
        template,
        jipdir,
        erase,
        resample,
        interleaved,
        sliceorder,
        realign_dof,
        realign_to_vol,
        warp,
        warp_njobs,
        warp_index,
        warp_file,
        warp_restrict,
        delta_te,
        dwell_time,
        manufacturer,
        blip_files,
        blip_enc_dirs,
        unwarp_direction,
        phase_file,
        magnitude_file,
        anatorient,
        funcorient,
        kernel_size,
        fslconfig,
        normalization_trf,
        coregistration_trf,
        recon1,
        recon2,
        auto,
        verbose):
    """ fMRI preprocessings using FSL, SPM, JIP, and ANTS.
    """
    # TODO: remove when all controls available in pypipe
    if not isinstance(erase, bool):
        erase = eval(erase)
        resample = eval(resample)
        interleaved = eval(interleaved)
        realign_to_vol = eval(realign_to_vol)
        warp = eval(warp)
        recon1 = eval(recon1)
        recon2 = eval(recon2)
        auto = eval(auto)
        warp_restrict = eval(warp_restrict)
        blip_files = None if blip_files == "" else eval(blip_files)
        blip_enc_dirs = eval(blip_enc_dirs)

    # Read input parameters
    realign_to_mean = not realign_to_vol
    subjdir = os.path.join(outdir, sid)
    cachedir = os.path.join(subjdir, "cachedir")
    outputs = {}
    if erase and os.path.isdir(subjdir):
        shutil.rmtree(subjdir)
    if not os.path.isdir(cachedir):
        os.makedirs(cachedir)
    nipype_memory = NipypeMemory(cachedir)
    joblib_memory = JoblibMemory(cachedir, verbose=verbose)
    def display_outputs(outputs, verbose, **kwargs):
        """ Simple function to display/store step outputs.
        """
        if verbose > 0:
            print("-" * 50)
            for key, val in kwargs.items():
                print("{0}: {1}".format(key, val))
            print("-" * 50)
        outputs.update(kwargs)

    # Check input parameters
    template_axes = guess_orientation(template)
    if template_axes != "RAS":
        raise ValueError("The template orientation must be 'RAS', '{0}' "
                         "found.".format(template_axes))
    check_jip_install(jipdir)
    if sliceorder not in ("ascending", "descending"):
        raise ValueError("Supported slice order are: ascending & descending.")

    # Slice timing
    fslenv = environment(fslconfig)
    if (fslenv["FSLDIR"] != os.environ.get("FSLDIR", "")):
        os.environ = concat_environment(os.environ, fslenv)
    st_dir = os.path.join(subjdir, STEPS["slice_timing"])
    if not os.path.isdir(st_dir):
        os.mkdir(st_dir)
    fsl.FSLCommand.set_default_output_type('NIFTI_GZ')
    interface = nipype_memory.cache(fsl.SliceTimer)
    returncode = interface(
        in_file=funcfile,
        interleaved=interleaved,
        slice_direction=3,
        time_repetition=repetitiontime,
        index_dir=False if sliceorder=="ascending" else True,
        out_file=os.path.join(
            st_dir, os.path.basename(funcfile).split(".")[0] + ".nii.gz"))
    st_outputs = returncode.outputs.get()
    slice_time_corrected_file = st_outputs["slice_time_corrected_file"]
    display_outputs(
        outputs, verbose, slice_time_corrected_file=slice_time_corrected_file)

    # B0 inhomogeneities: topup or fugue or None + motion induced
    if warp:
        warp_dir = os.path.join(subjdir, STEPS["warp"])
        if not os.path.isdir(warp_dir):
            os.mkdir(warp_dir)
        if blip_files is not None:
            interface = joblib_memory.cache(topup)
            fieldmap_hz_file, unwarped_epi_file = interface(
                blip_up_file=blip_files[0],
                blip_down_file=blip_files[1],
                blip_up_phase_enc_dir=blip_enc_dirs[0],
                blip_down_phase_enc_dir=blip_enc_dirs[1],
                apply_to=slice_time_corrected_file,
                unwarp_direction=unwarp_direction,
                dwell_time=dwell_time,
                outdir=warp_dir,
                fsl_sh=fslconfig)
        elif phase_file is not None:
            interface = joblib_memory.cache(fugue)
            magnitude_brain_mask_file, vsm_file, unwarped_epi_file = interface(
                epi_file=slice_time_corrected_file,
                phase_file=phase_file,
                magnitude_file=magnitude_file,
                delta_te=delta_te,
                dwell_time=dwell_time,
                unwarp_direction=unwarp_direction,
                manufacturer=manufacturer,
                outdir=warp_dir,
                fsl_sh=fslconfig,
                verbose=verbose)
        else:
            unwarped_epi_file = slice_time_corrected_file
        interface = joblib_memory.cache(timeserie_to_reference)
        b0_corrected_file = interface(
            tfile=unwarped_epi_file,
            rindex=warp_index,
            restrict_deformation=warp_restrict,
            rfile=warp_file,
            outdir=warp_dir,
            njobs=warp_njobs,
            clean_tmp=True)
    else:
        b0_corrected_file = slice_time_corrected_file
    display_outputs(
        outputs, verbose, b0_corrected_file=b0_corrected_file)

    # Reorient images not in RAS coordinate system and
    # reorient images to match the orientation of the standard MNI152 template.
    reorient_dir = os.path.join(subjdir, STEPS["reorient"])
    if not os.path.isdir(reorient_dir):
        os.mkdir(reorient_dir)
    reoriented_funcfile = b0_corrected_file
    reoriented_anatfile = anatfile
    interface = joblib_memory.cache(reorient_image)
    if funcorient != "RAS":
        reoriented_funcfile = interface(
            b0_corrected_file,
            axes=funcorient,
            prefix="o",
            output_directory=reorient_dir)
    if anatorient != "RAS":
        reoriented_anatfile = interface(
            anatfile,
            axes=anatorient,
            prefix="o",
            output_directory=reorient_dir)
    standard_funcfile = os.path.join(
        reorient_dir, "d" + os.path.basename(reoriented_funcfile).split(".")[0])
    standard_anatfile = os.path.join(
        reorient_dir, "d" + os.path.basename(reoriented_anatfile).split(".")[0])
    interface = joblib_memory.cache(fslreorient2std)
    standard_funcfile = interface(
        reoriented_funcfile,
        standard_funcfile,
        fslconfig=fslconfig)
    standard_anatfile = interface(
        reoriented_anatfile,
        standard_anatfile,
        fslconfig=fslconfig)
    display_outputs(
        outputs, verbose, standard_funcfile=standard_funcfile,
        standard_anatfile=standard_anatfile)

    # Downsample template
    if resample:
        template = resample_image(
            source_file=template,
            target_file=standard_anatfile,
            out_file=os.path.join(subjdir, "template.nii.gz"),
            fslconfig=fslconfig)

    # Realign
    realign_dir = os.path.join(subjdir, STEPS["realign"])
    if not os.path.isdir(realign_dir):
        os.mkdir(realign_dir)
    realign_func_rootfile = os.path.join(
        realign_dir, "r" + os.path.basename(standard_funcfile).split(".")[0])
    interface = joblib_memory.cache(mcflirt)
    realign_funcfile, realign_func_meanfile, realign_func_parfile = interface(
        in_file=standard_funcfile,
        out_fileroot=realign_func_rootfile,
        cost="normcorr",
        bins=256,
        dof=realign_dof,
        refvol=warp_index,
        reffile=warp_file,
        reg_to_mean=realign_to_mean,
        mats=True,
        plots=True,
        verbose=verbose,
        shfile=fslconfig)
    display_outputs(
        outputs, verbose, realign_funcfile=realign_funcfile,
        realign_func_meanfile=realign_func_meanfile)

    # Early stop detected
    if recon1:
        print("[warn] User requested a processing early stop. Remove the 'recon1' "
              "option to resume.")
        return outputs

    # Normalization.
    normalization_dir = os.path.join(subjdir, STEPS["normalization"])
    if not os.path.isdir(normalization_dir):
        os.mkdir(normalization_dir)
    if normalization_trf is not None:
        shutil.copy(normalization_trf, normalization_dir)
    interface = joblib_memory.cache(jip_align)
    (register_anatfile, register_anat_maskfile,
     native_anat_maskfile, align_normfile) = interface(
        source_file=standard_anatfile,
        target_file=template,
        outdir=normalization_dir,
        jipdir=jipdir,
        prefix="w",
        auto=auto,
        non_linear=True,
        fslconfig=fslconfig)
    display_outputs(
        outputs, verbose, register_anatfile=register_anatfile,
        register_anat_maskfile=register_anat_maskfile,
        native_anat_maskfile=native_anat_maskfile, align_normfile=align_normfile)

    # Tissues segmentation and spatial intensity variations correction.
    inhomogeneities_dir = os.path.join(subjdir, STEPS["inhomogeneities"])
    if not os.path.isdir(inhomogeneities_dir):
        os.mkdir(inhomogeneities_dir)
    biascorrected_anatfile = os.path.join(
        inhomogeneities_dir, "n4_" + os.path.basename(native_anat_maskfile))
    bias_anatfile = os.path.join(
        inhomogeneities_dir, "n4field_" + os.path.basename(native_anat_maskfile))
    n4 = ants.N4BiasFieldCorrection()
    interface = nipype_memory.cache(ants.N4BiasFieldCorrection)
    returncode = interface(
        dimension=3,
        input_image=native_anat_maskfile,
        bspline_fitting_distance=200,
        shrink_factor=2,
        n_iterations=[50,50,40,30],
        convergence_threshold=1e-6,
        output_image=biascorrected_anatfile,
        bias_image=bias_anatfile)
    n4_outputs = returncode.outputs.get()
    display_outputs(
        outputs, verbose, bias_anatfile=bias_anatfile,
        biascorrected_anatfile=biascorrected_anatfile)

    # Coregistration.
    coregistration_dir = os.path.join(subjdir, STEPS["coregistration"])
    if not os.path.isdir(coregistration_dir):
        os.mkdir(coregistration_dir)
    if coregistration_trf is not None:
        shutil.copy(coregistration_trf, coregistration_dir)
    interface = joblib_memory.cache(jip_align)
    (register_func_meanfile, register_func_mean_maskfile,
     native_func_mean_maskfile, align_coregfile) = interface(
        source_file=realign_func_meanfile,
        target_file=biascorrected_anatfile,
        outdir=coregistration_dir,
        jipdir=jipdir,
        prefix="w",
        auto=auto,
        non_linear=False,
        fslconfig=fslconfig)
    display_outputs(
        outputs, verbose, register_func_meanfile=register_func_meanfile,
        register_func_mean_maskfile=register_func_mean_maskfile,
        native_func_mean_maskfile=native_func_mean_maskfile,
        align_coregfile=align_coregfile)

    # Early stop detected
    if recon2:
        print("[warn] User requested a processing early stop. Remove the 'recon2' "
              "option to resume.")
        return outputs

    # Wrap functional: resample the functional serie and mask the registered serie.
    wrap_dir = os.path.join(subjdir, STEPS["wrap"])
    if not os.path.isdir(wrap_dir):
        os.mkdir(wrap_dir)
    interface = joblib_memory.cache(apply_jip_align)
    deformed_files = interface(
        apply_to_files=[realign_funcfile],
        align_with=[align_coregfile, align_normfile],
        outdir=wrap_dir,
        jipdir=jipdir,
        prefix="w",
        apply_inv=False)
    register_funcfile = deformed_files[0]
    register_func_mask_fileroot = os.path.join(
        wrap_dir, "m" + os.path.basename(register_funcfile).split(".")[0])
    interface = joblib_memory.cache(apply_mask)
    register_func_maskfile = interface(
        input_file=register_funcfile,
        output_fileroot=register_func_mask_fileroot,
        mask_file=template,
        fslconfig=fslconfig)
    display_outputs(
        outputs, verbose, register_funcfile=register_funcfile,
        register_func_maskfile=register_func_maskfile)

    # Smooth the functional serie.
    smooth_dir = os.path.join(subjdir, STEPS["smooth"])
    if not os.path.isdir(smooth_dir):
        os.mkdir(smooth_dir)
    interface = nipype_memory.cache(fsl.Smooth)
    returncode = interface(
        in_file=register_func_maskfile,
        fwhm=kernel_size,
        output_type="NIFTI",
        smoothed_file=os.path.join(
            smooth_dir,
            "smooth_" + os.path.basename(register_func_maskfile).split(".")[0] +
            ".nii"))
    smooth_outputs = returncode.outputs.get()
    smoothed_file = smooth_outputs["smoothed_file"]
    display_outputs(outputs, verbose, smoothed_file=smoothed_file)

    # Copy the results to the root directory: use Nifti format.
    nibabel.load(smoothed_file).to_filename(
        os.path.join(subjdir, "sMNI.nii"))
    nibabel.load(register_func_maskfile).to_filename(
        os.path.join(subjdir, "MNI.nii"))
    nibabel.load(register_anat_maskfile).to_filename(
        os.path.join(subjdir, "anat.nii"))

    # Compute some snaps assessing the different processing steps.
    snapdir = os.path.join(subjdir, STEPS["snaps"])
    if not os.path.isdir(snapdir):
        os.mkdir(snapdir)
    interface = joblib_memory.cache(triplanar)
    # > generate coregistration plot
    coregister_fileroot = os.path.join(snapdir, "coregister")
    coregister_file = interface(
        input_file=register_func_meanfile,
        output_fileroot=coregister_fileroot,
        overlays=[standard_anatfile],
        overlays_colors=None,
        contours=True,
        edges=False,
        overlay_opacities=[0.7],
        resolution=300)
    # > generate normalization plot
    normalize_fileroot = os.path.join(snapdir, "normalization")
    normalize_file = interface(
        input_file=register_anatfile,
        output_fileroot=normalize_fileroot,
        overlays=[template],
        overlays_colors=None,
        contours=True,
        edges=False,
        overlay_opacities=[0.7],
        resolution=300)
    # > generate a motion parameter plot
    interface = joblib_memory.cache(plot_fsl_motion_parameters)
    realign_motion_file = os.path.join(snapdir, "realign_motion_parameters.png")
    interface(realign_func_parfile, realign_motion_file)
    display_outputs(
        outputs, verbose, normalize_file=normalize_file,
        realign_motion_file=realign_motion_file, coregister_file=coregister_file)

    # Generate a QC reporting
    reportdir = os.path.join(subjdir, STEPS["report"])
    reportfile = os.path.join(reportdir, "QC_preproc_{0}.pdf".format(sid))
    if not os.path.isdir(reportdir):
        os.mkdir(reportdir)
    interface = joblib_memory.cache(generate_pdf)
    tic = datetime.now()
    date = "{0}-{1}-{2}".format(tic.year, tic.month, tic.day)
    interface(
        datapath=snapdir,
        struct_file=os.path.join(
            os.path.abspath(os.path.dirname(pypreclin.__file__)), "utils",
            "resources", "pypreclin_qcpreproc.json"),
        author="NeuroSpin",
        client="-",
        poweredby="FSL-SPM-Nipype-JIP",
        project="-",
        timepoint="-",
        subject=sid,
        date=date,
        title="fMRI Preprocessing QC Reporting",
        filename=reportfile,
        pagesize=None,
        left_margin=10,
        right_margin=10,
        top_margin=20,
        bottom_margin=20,
        show_boundary=False,
        verbose=0)
    display_outputs(outputs, verbose, reportfile=reportfile)

    return outputs


def preproc_multi(
        bidsdir, template, jipdir, fslconfig, auto, resample, anatorient="RAS",
        funcorient="RAS", njobs=1, simage=None, shome=None, sbinds=None,
        verbose=0):
    """ Perform the FMRI preprocessing on a BIDS organized directory in
    parallel (without FUGUE or TOPUP).
    This function can be called with a singularity image that contains all the
    required software.

    If a 'jip_trf' directory is available in the session directory, the code
    will use the available JIP transformation.

    Parameters
    
    bidsdir: str
        the BIDS organized directory.
    template: str
        the path to the template in RAS coordiante system.
    jipdir: str
        the jip software binary path.
    fslconfig: str
        the FSL configuration script.
    auto: bool
        control the JIP window with the script.
    resample: bool
        if set resample the input template to fit the anatomical image.
    anatorient: str, default "RAS"
        the input anatomical image orientation.
    funcorient: str, default "RAS"
        the input functional image orientation.
    njobs: int, default 1
        the number of parallel jobs.
    simage: simg, default None
        a singularity image.
    shome: str, default None
        a home directory for the singularity image.
    sbinds: str, default None
        shared directories for the singularity image.
    verbose: int, default 0
        control the verbosity level.
    """
    # TODO: remove when all controls available in pypipe
    if not isinstance(auto, bool):
        auto = eval(auto)
        resample = eval(resample)
        sbinds = eval(sbinds)

    # Parse the BIDS directory
    desc_file = os.path.join(bidsdir, "dataset_description.json")
    if not os.path.isfile(desc_file):
        raise ValueError(
            "Expect '{0}' in a BIDS organized directory.".format(desc_file))
    with open(desc_file, "rt") as of:
        desc = json.load(of)
    name = desc["Name"]
    if verbose > 0:
        print("Processing dataset '{0}'...".format(name))
    dataset = {name: {}}
    funcfiles, anatfiles, subjects, outputs, trs, encdirs, normfiles, coregfiles = [
        [] for _ in range(8)]
    outdir = os.path.join(bidsdir, "derivatives", "preproc")
    for sesdir in glob.glob(os.path.join(bidsdir, "sub-*", "ses-*")):
        split = sesdir.split(os.sep)
        sid, ses = split[-2:]
        _anatfiles = glob.glob(os.path.join(sesdir, "anat", "sub-*T1w.nii.gz"))
        _funcfiles = glob.glob(os.path.join(sesdir, "func", "sub-*bold.nii.gz"))
        if len(_anatfiles) != 1:
            if verbose > 0:
                print("Skip session '{0}': no valid anat.".format(sesdir))
            continue
        if len(_funcfiles) == 0:
            if verbose > 0:
                print("Skip session '{0}': no valid func.".format(sesdir))
            continue
        descs = []
        for path in _funcfiles:
            runs = re.findall(".*_(run-[0-9]*)_.*", path)
            if len(runs) != 1:
                if verbose > 0:
                    print("Skip path '{0}': not valid name.".format(path))
                continue
            sesdir = os.path.join(outdir, sid, ses)
            rundir = os.path.join(sesdir, runs[0])
            if not os.path.isdir(rundir):
                os.makedirs(rundir)
            desc_file = path.replace(".nii.gz", ".json")
            if not os.path.isfile(desc_file):
                break
            with open(desc_file, "rt") as of:
                _desc = json.load(of)
            if _desc["PhaseEncodingDirection"].startswith("i"):
                warp_restrict = [1, 0, 0]
            elif _desc["PhaseEncodingDirection"].startswith("j"):
                warp_restrict = [0, 1, 0]
            elif _desc["PhaseEncodingDirection"].startswith("k"):
                warp_restrict = [0, 0, 1]
            else:
                raise ValueError(
                    "Unknown encode phase direction : {0}...".format(
                        _desc["PhaseEncodingDirection"]))
            jip_normalization = os.path.join(
                sesdir, "jip_trf", "Normalization", "align.com")
            if not os.path.isfile(jip_normalization):
                if verbose > 0:
                    print("No JIP normalization align.com file found "
                          "in {0}.".format(sesdir))
                jip_normalization = None
            jip_coregistration = os.path.join(
                sesdir, "jip_trf", "Coregistration", "align.com")
            if not os.path.isfile(jip_coregistration):
                if verbose > 0:
                    print("No JIP coregistration align.com file found "
                          "in {0}.".format(sesdir))
                jip_coregistration = None
            desc = {
                "tr": _desc["RepetitionTime"],
                "warp_restrict": warp_restrict,
                "output": rundir}
            descs.append(desc)
            funcfiles.append(path)
            anatfiles.append(_anatfiles[0])
            subjects.append(sid)
            outputs.append(rundir)
            trs.append(_desc["RepetitionTime"])
            encdirs.append(warp_restrict)
            normfiles.append(jip_normalization)
            coregfiles.append(jip_coregistration)
        if len(_funcfiles) != len(descs):
            if verbose > 0:
                print("Skip session '{0}': no valid desc.".format(sesdir))
            continue
        if sid not in dataset[name]:
            dataset[name][sid] = {}
        dataset[name][sid][ses] = {
            "anat": _anatfiles[0],
            "func": _funcfiles,
            "desc": descs}
    if verbose > 0:
        pprint(dataset)

    # Preprare inputs
    expected_size = len(subjects)
    if expected_size == 0:
        if verbose > 0:
            print("no data to process.")
        return None
    for name, item in (("outputs", outputs), ("funcfiles", funcfiles),
                       ("anatfiles", anatfiles), ("subjects", subjects),
                       ("trs", trs), ("encdirs", encdirs),
                       ("normfiles", normfiles), ("coregfiles", coregfiles)):
        if item is None:
            continue
        if verbose > 0:
            print("[{0}] {1} : {2} ... {3}".format(
                name.upper(), len(item), item[0], item[-1]))
        if len(item) != expected_size:
            raise ValueError("BIDS dataset not aligned.")

    # Run preproc
    scriptdir = os.path.join(os.path.dirname(pypreclin.__file__), "scripts")
    script = os.path.join(scriptdir, "pypreclin_preproc_fmri")
    python_cmd = "python3"
    if not os.path.isfile(script):
        script = "pypreclin_preproc_fmri"
        python_cmd = None
    if simage is not None:
        script = "singularity run --home {0} ".format(
            shome or tempfile.gettempdir())
        sbinds = sbinds or []
        for path in sbinds:
            script += "--bind {0} ".format(path)
        script += simage
        python_cmd = None
    logdir = os.path.join(bidsdir, "derivatives", "logs")
    if not os.path.isdir(logdir):
        os.makedirs(logdir)
    logfile = os.path.join(
        logdir, "pypreclin_{0}.log".format(datetime.now().isoformat()))
    status, exitcodes = hopla(
        script,
        hopla_iterative_kwargs=["f", "a", "s", "o", "r", "WR", "N", "M"],
        f=funcfiles,
        a=anatfiles,
        s=subjects,
        o=outputs,
        r=trs,
        t=template,
        j=jipdir,
        resample=resample,
        W=True,
        WN=1,
        WR=encdirs,
        NA=anatorient,
        NF=funcorient,
        C=fslconfig,
        N=normfiles,
        M=coregfiles,
        A=auto,
        hopla_python_cmd=python_cmd,
        hopla_use_subprocess=True,
        hopla_verbose=1,
        hopla_cpus=njobs,
        hopla_logfile=logfile)

    return {"logfile": logfile}
    



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
B0 inhomogeneties correction.
"""

# Sytem import
import os
import shutil
import subprocess

# Package import
import pypreclin
from pypreclin.utils.export import ungzip_file
from pypreclin.utils.filetools import average_timeserie

# Third party import
from pyconnectome.wrapper import FSLWrapper
from pyconnectome.utils.segtools import bet2
from pyconnectome.utils.preproctools import topup as _topup
from pyconnectome.utils.preproctools import fsl_prepare_fieldmap
from pyconnectome.configuration import environment, concat_environment
from nipype.interfaces.fsl.preprocess import FUGUE


DIR_MAP = {
    "i": "x",
    "i-": "x-",
    "j": "y",
    "j-": "y-",
    "k": "z",
    "k-": "z-"
}


def mc_undist(funcfile, outdir, spmpath, fslpath):
    """ Slice by slice motion correction: deprecated
    """
    # Extract the brain with FSL bet
    outputs = bet2(funcfile, outdir, outline=False, mask=True,
                   skull=False, nooutput=True, f=0.15, g=0, radius=None,
                   smooth=None, c=None, threshold=False, mesh=False,
                   shfile=fslpath)
    gzmaskfile = outputs[1].replace("_mask", "_msk")
    os.rename(outputs[1], gzmaskfile)
    maskfile = ungzip_file(gzmaskfile, prefix="")
    os.remove(gzmaskfile)

    # Generate the matlab job
    shutil.copy(funcfile, outdir)
    jobfile = os.path.join(
        os.path.dirname(pypreclin.__file__), "preproc", "resources", "leuven",
                        "mc_undist_wouterdante_tasserie_job.m")
    info = {
        "SPMPATH": spmpath,
        "PREPROCPATH": os.path.join(os.path.dirname(pypreclin.__file__),
                                    "preproc"),
        "CWD": outdir + os.path.sep,
        "BASENAME": os.path.basename(funcfile)
    }
    with open(jobfile, "rt") as open_file:
        job = open_file.read()
    jobfile = os.path.join(outdir, "mc_undist_job.m")
    with open(jobfile, "wt") as open_file:
        open_file.write(job.format(**info))

    # Call the undist script
    cmd = ["matlab", "-nodisplay", "-nosplash", "-nodesktop", "-r",
           "run {0}".format(jobfile)]
    subprocess.check_call(cmd)
    undist_file = os.path.join(outdir, "u_" + os.path.basename(funcfile))

    return undist_file


def topup(blip_up_file, blip_down_file, blip_up_phase_enc_dir,
          blip_down_phase_enc_dir, outdir, fsl_sh, apply_to=None,
          unwarp_direction=None, dwell_time=None, verbose=0):
    """ Bias field correction using topup.

    Note: we do not set the total readout times for both acquisitons, and we
    expect that the readout time is identical for all acquisitions. This 
    parameter simply serves to ensure that the estimated field is correctly
    scaled in Hz, but it doesn't affect the result in terms of undistorted
    images. We set it to 1, meaning that the field will be scaled as 'voxels 
    displacement' instead.

    Parameters
    ----------
    blip_up_file:  str
        path to EPI file acquired in opposite phase enc. directions: blip up.
    blip_down_file:  str
        path to EPI file acquired in opposite phase enc. directions: blip down.
    blip_up_phase_enc_dir: str
        the phase enc. direction for the blip up.
    blip_down_phase_enc_dir: str
        the phase enc. direction for the blip down.
    phase_enc_dirs: list of str
        the phase enc. directions.
    outdir: str
        directory for outputs.
    fsl_sh: str
        path to fsl setup sh file.
    apply_to: str, default None
        apply the topup correction to the EPI volume.
    unwarp_direction: str, default None
        apply topup correction in this direction.
    dwell_time: float, default None
        set the EPI dwell time per phase-encode line, same as echo spacing (ms).
    verbose: int, default 0
        control the verbosity level.

    Returns
    -------
    fieldmap_hz_file: str
        the fieldmap in Hz
    unwarped_epi_file: str
        unwarped EPI file.
    """
    # Call topup
    fieldmap_hz_file, _, _ = _topup(
        b0s=[blip_up_file, blip_down_file],
        phase_enc_dirs=[blip_up_phase_enc_dir, blip_down_phase_enc_dir],
        readout_time=1,
        outroot=outdir,
        apply_to=None,
        fsl_sh=fsl_sh)

    # Convert the fieldmap in rad/s to Hz
    fieldmap_file = fieldmap_hz_file.replace(".nii.gz", "_rads.nii.gz")
    cmd = ["fslmaths", fieldmap_hz_file, "-mul", "6.28", fieldmap_file]
    fslprocess = FSLWrapper(shfile=fsl_sh)
    fslprocess(cmd=cmd)

    # Update env
    fslenv = environment(fsl_sh)
    if (fslenv["FSLDIR"] != os.environ.get("FSLDIR", "")):
        os.environ = concat_environment(os.environ, fslenv)

    # Apply topup
    # Unwarping an input image (fieldmap in rad/s is known)
    unwarp_direction = DIR_MAP[unwarp_direction]
    unwarped_epi_file = os.path.join(outdir, "unwarped_epi_file.nii.gz")
    fugue = FUGUE()
    fugue.inputs.in_file = apply_to
    fugue.inputs.fmap_in_file = fieldmap_file
    fugue.inputs.unwarp_direction = unwarp_direction
    fugue.inputs.unwarped_file = unwarped_epi_file
    fugue.inputs.dwell_time = dwell_time * 1e-3
    fugue.inputs.output_type = "NIFTI_GZ"
    if verbose > 0:
        print(fugue.cmdline)
    returncode = fugue.run()

    return fieldmap_hz_file, unwarped_epi_file


def fugue(epi_file, phase_file, magnitude_file, delta_te, dwell_time,
          unwarp_direction, manufacturer, outdir, fsl_sh, verbose=0):
    """ Unwarping of an EPI image based on fieldmap data using fugue.

    Note: Brain extraction of the EPI image is very important and must be
    tight - that is, it must exclude all non-brain voxels and any voxels with
    only a small partial volume contribution. The reason for this is that
    these areas are normally very noisy in the phase image. The exclusion
    of brain voxels is actually fine and will have no repercussions, since the
    fieldmap is extrapolated beyond this mask.

    Note: If parallel acceleration is used in the EPI acquisition then the
    *effective* echo spacing (dwell time) is the actual echo spacing between
    acquired lines in k-space divided by the acceleration factor.

    Parameters
    ----------
    epi_file: str
        the EPI file to unwarp.
    phase_file: str
        the phase image in the EPI space.
    magnitude_file: str
        the magnitude image in the EPI space.
    delta_te: float
        the echo time difference of the fieldmap sequence - find this out form
        the operator (defaults are *usually* 2.46ms on SIEMENS).
    dwell_time: float
        set the EPI dwell time per phase-encode line, same as echo spacing (ms).
    unwarp_direction: str
        specifies direction of warping: ('x' or 'y' or 'z' or 'x-' or 'y-'
        or 'z-')
    manufacturer: str
        must be SIEMENS.
    outdir: str
        directory for outputs.
    fsl_sh: str
        path to fsl setup sh file.
    verbose: int, default 0
        control the verbosity level.

    Returns
    -------
    magnitude_brain_mask_file: str
        the brain mask.
    vsm_file: str
        voxel shift map file.
    unwarped_epi_file: str
        unwarped EPI file.
    """
    # Extract brain
    avgmagnitude_file = os.path.join(outdir, "avgmag.nii.gz")
    average_timeserie(magnitude_file, avgmagnitude_file, fslconfig=fsl_sh)
    outputs = bet2(avgmagnitude_file, outdir, mask=True, f=0.35, shfile=fsl_sh)
    #outputs = bet2(outputs[0], outdir, mask=True, f=0.6, shfile=fsl_sh)
    magnitude_brain_file, magnitude_brain_mask_file = outputs[:2]

    # Update env
    fslenv = environment(fsl_sh)
    if (fslenv["FSLDIR"] != os.environ.get("FSLDIR", "")):
        os.environ = concat_environment(os.environ, fslenv)

    # Prepare a rad/s fieldmap
    fieldmap_file = os.path.join(outdir, "fieldmap.nii.gz")
    fieldmap_file, fieldmap_hz_file = fsl_prepare_fieldmap(
        manufacturer=manufacturer.upper(),
        phase_file=phase_file,
        brain_magnitude_file=magnitude_brain_file,
        output_file=fieldmap_file,
        delta_te=str(delta_te),
        fsl_sh=fsl_sh)

    # Computing the vsm (unwrapped phase map is known)
    unwarp_direction = DIR_MAP[unwarp_direction]
    vsm_file = os.path.join(outdir, "vsm.nii.gz")
    fugue = FUGUE()
    fugue.inputs.fmap_in_file = fieldmap_file
    fugue.inputs.mask_file = magnitude_brain_mask_file
    fugue.inputs.dwell_time = dwell_time * 1e-3
    fugue.inputs.asym_se_time = delta_te * 1e-3
    fugue.inputs.unwarp_direction = unwarp_direction
    fugue.inputs.save_unmasked_shift = True
    fugue.inputs.shift_out_file = vsm_file
    fugue.inputs.smooth2d = 2
    fugue.inputs.output_type = "NIFTI_GZ"
    fugue.environ = {}
    if verbose > 0:
        print(fugue.cmdline)
    returncode = fugue.run()
    #outputs = returncode.outputs.get()
    #vsm_file = outputs["shift_out_file"]

    # Unwarping an input image (shift map is known)
    unwarped_epi_file = os.path.join(outdir, "unwarped_epi_file.nii.gz")
    fugue = FUGUE()
    fugue.inputs.in_file = epi_file
    fugue.inputs.mask_file = magnitude_brain_mask_file
    fugue.inputs.shift_in_file = vsm_file
    fugue.inputs.unwarp_direction = unwarp_direction
    fugue.inputs.unwarped_file = unwarped_epi_file
    fugue.inputs.output_type = "NIFTI_GZ"
    if verbose > 0:
        print(fugue.cmdline)
    returncode = fugue.run()
    #outputs = returncode.outputs.get()
    #unwarped_epi_file = outputs["unwarped_file"]

    return magnitude_brain_mask_file, vsm_file, unwarped_epi_file


        



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2018
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################


"""
deprecated
"""

# System import
import subprocess
import fcntl
import os
import sys
import time

# Third party import
from filelock import FileLock


def non_block_read(output):
    """ Function to read a stream without blocking.
    """
    fd = output.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
    try:
        return output.read()
    except:
        return ""


def search_align_windows():
    """ List the windows called align.
    
    Returns
    -------
    align_windows: list
        the align windows ids.
    """
    cmd = ["xdotool", "search", "--name", "align"]
    process = subprocess.Popen(
        cmd, env=os.environ, stdout=subprocess.PIPE, universal_newlines=True)
    align_windows = process.communicate()[0].split("\n")
    return align_windows


def send_key(key, win_id):
    """ Send a keyboard key to a windows.

    Parameters
    ----------
    key: str
        the keyboard event to be sent.
    win_id: int
        the window id to control.
    """
    cmd = ["xdotool", "windowactivate", "--sync", str(win_id), "key", key]
    subprocess.check_call(cmd)


def create_win(cmd, lock_file, default_windows):
    """ Create a JIP windows and return his id.

    Parameters
    ----------
    cmd: list of str
        the JIP command that will popup a window.
    lock_file: str
        the path to a file used to acquired a lock in order to identify
        properly the window id.
    default_windows: list
        the default windows ID.

    Returns
    -------
    process: object
        the JIP process.
    win_id: int
        the associated window id.
    """
    with FileLock(lock_file):
        process = subprocess.Popen(
                cmd, env=os.environ, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, universal_newlines=True)
        win_id = None
        while True:
            time.sleep(2)
            if win_id is None:
                new_windows = search_align_windows()
                active_windows = list(set(new_windows) - set(default_windows))
                assert len(active_windows) <= 1, "Parallel jobs not supported yet."
                if len(active_windows) == 1:
                    win_id = active_windows[0]
                    break
    return process, win_id


def call_auto(cmd, lock_file, refresh_winid=True):
    """ Execute a JIP command and control the graphical interface.

    Parameters
    ----------
    cmd: list of str
        the JIP command that will popup a window.
    lock_file: str
        the path to a file used to acquired a lock in order to identify
        properly the window id.
    refresh_winid: bool (optional, default True)
        if set check that only one JIP alignment is performed at a time. The
        windows ID is refreshed at each iteration.

    Returns
    -------
    mi: float
        the final optimizer cost function value (mutual information).
    """
    default_windows = search_align_windows()
    process, win_id = create_win(cmd, lock_file, default_windows)
    send_run = False
    while True:
        time.sleep(2)          
        if not send_run:
            send_key("a", win_id)
            send_run = True
        line = non_block_read(process.stdout)
        if line == "":
            continue
        if refresh_winid:
            new_windows = search_align_windows()
            active_windows = list(set(new_windows) - set(default_windows))
            assert len(active_windows) == 1, "Parallel jobs not supported yet."
            if active_windows[0] != win_id:
                print("Refreshing JIP windows ID {0} -> {1}.".format(
                    win_id, active_windows[0]))
                win_id = active_windows[0]
        if "MI =" in line:
            mi = float(line.split("=")[1])
            break
    send_key("q", win_id)
    send_key("y", win_id)
    stdout, stderr = process.communicate()

    return mi



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Registration utilities.
"""

# System import
import os
import subprocess
import nibabel
import numpy
import shutil
import multiprocessing

# Hopla import
from hopla.converter import hopla

# Package import
from pypreclin.utils.reorient import check_orientation
from pypreclin.utils.export import ungzip_file
from pypreclin.utils.export import gzip_file
from pypreclin.utils.reorient import swap_affine
from pypreclin import preproc

# Third party import
import pyconnectome
from pyconnectome.utils.filetools import apply_mask
from pyconnectome import DEFAULT_FSL_PATH
from pyconnectome.utils.regtools import flirt
from nipype.interfaces.ants import AffineInitializer
from nipype.interfaces.ants import ApplyTransforms
from transforms3d.affines import decompose44
from transforms3d.euler import mat2euler
from scipy.io import loadmat
import nibabel


def timeserie_to_reference(tfile, outdir, rindex=None,
                           restrict_deformation=(1, 1, 1), rfile=None, njobs=1,
                           clean_tmp=True):
    """ Register all the fMRI volumes to a reference volume identified by his
    index in the timeserie.

    The registration used is a non-linear regisration.

    Parameters
    ----------
    tfile: str
        the file that containes the timeserie.
    outdir: str
        the destination folder.
    rindex: int, default None
        the reference volume index in the timeserie.
    restrict_deformation: 3-uplet
        restrict the deformation in the given axis.
    rfile: str, default None
        the reference volume.
    njobs: int, default 1
        the desired number of parallel job during the registration.
    clean_tmp: bool, default True
        if set, clean the temporary results.

    Returns
    -------
    resfile: str
        the registration result.
    """
    # Check input index
    im = nibabel.load(tfile)
    array = im.get_data()
    if array.ndim != 4:
        raise ValueError("A timeserie (4d volume) is expected.")
    reference_image = None
    if rindex is not None:
        if rindex < 0 or rindex > array.shape[3]:
            raise ValueError(
                "Index '{0}' is out of bound considering the last dimension "
                "as the time dimension '{1}'.".format(rindex, array.shape))
    elif rfile is not None:
        reference_image = rfile
    else:
        raise ValueError("You need to specify a reference file or index.")

    # Split the timeserie
    tmpdir = os.path.join(outdir, "tmp")
    if not os.path.isdir(tmpdir):
        os.mkdir(tmpdir)
    moving_images = []
    outdirs = []
    for i in range(array.shape[3]):
        _im = nibabel.Nifti1Image(array[..., i], im.affine)
        _outfile = os.path.join(tmpdir, str(i).zfill(4) + ".nii.gz")
        outdirs.append(os.path.join(tmpdir, str(i).zfill(4)))
        if not os.path.isdir(outdirs[-1]):
            os.mkdir(outdirs[-1])
        nibabel.save(_im, _outfile)
        moving_images.append(_outfile)
        if reference_image is None and i == rindex:
            reference_image = _outfile

    # Start volume to volume non rigid registration
    scriptdir = os.path.join(os.path.dirname(pyconnectome.__file__), "scripts")
    script = os.path.join(scriptdir, "pyconnectome_ants_register")
    python_cmd = "python3"
    if not os.path.isfile(script):
        script = "pyconnectome_ants_register"
        python_cmd = None
    logfile = os.path.join(tmpdir, "log")
    if os.path.isfile(logfile):
        os.remove(logfile)
    status, exitcodes = hopla(
        script,
        b="/usr/lib/ants",
        o=outdirs,
        i=moving_images,
        r=reference_image,
        w=1,
        D=3,
        G=0.2,
        J=1,
        N=True,
        B=True,
        R=list(restrict_deformation),
        V=2,
        hopla_iterative_kwargs=["o", "i"],
        hopla_cpus=njobs,
        hopla_logfile=logfile,
        hopla_python_cmd=python_cmd,
        hopla_use_subprocess=True,
        hopla_verbose=1)
    if not (numpy.asarray(list(exitcodes.values())) == 0).all():
        raise ValueError("The registration failed, check the log "
                         "'{0}'.".format(logfile))

    # Start timeserie concatenation
    timeserie = []
    affine = None
    for path in outdirs:
        sid = os.path.basename(path)
        _im = nibabel.load(os.path.join(
            path,  "ants_2WarpToTemplate_{0}.nii.gz".format(sid)))
        if affine is None:
            affine = _im.affine
        elif not numpy.allclose(affine, im.affine, atol=1e-3):
            raise ValueError("Affine matrices must be the same: {0} - "
                             "{1}.".format(outdirs[0], path))
        data = _im.get_data()
        data.shape += (1, )
        timeserie.append(data)
    registered_array = numpy.concatenate(timeserie, axis=3)
    _im = nibabel.Nifti1Image(registered_array, affine)
    resfile = os.path.join(outdir, "ants_WarpToTemplate.nii.gz")
    nibabel.save(_im, resfile)

    # Clean temporary files if requested
    if clean_tmp:
        shutil.rmtree(tmpdir, ignore_errors=True)

    return resfile


def jip_align(source_file, target_file, outdir, jipdir, prefix="w",
              auto=False, non_linear=False, fslconfig=DEFAULT_FSL_PATH):
    """ Register a source image to a taget image using the 'jip_align'
    command.

    Parameters
    ----------
    source_file: str (mandatory)
        the source Nifti image.
    target_file: str (mandatory)
        the target Nifti masked image.
    outdir: str (mandatory)
        the destination folder.
    jipdir: str (mandatory)
        the jip binary path.
    prefix: str (optional, default 'w')
        prefix the generated file with this character.
    auto: bool (optional, default False)
        if set control the JIP window with the script.
    non_linear: bool (optional, default False)
        in the automatic mode, decide or not to compute the non-linear
        deformation field.
    fslconfig: str (optional)
        the FSL .sh configuration file.

    Returns
    -------
    register_file: str
        the registered image.
    register_mask_file: str
        the registered and masked image.
    native_masked_file: str
        the masked image in the native space.
    """
    # Check input image orientation: must be the same
    same_orient, orients = check_orientation([source_file, target_file])
    if not same_orient:
        raise ValueError(
            "Source file '{0}' ({2}) and taget file '{1}' ({3}) must have "
            "the same orientation for JIP to work properly.".format(
                source_file, target_file, orients[0], orients[1]))

    # Try to init the affine deformation
    # ToDo: find a way to init the JIP deformation
    if False:
        nb_cpus = multiprocessing.cpu_count()
        init_file = os.path.join(outdir, "init_affine.mat")
        init = AffineInitializer()
        init.inputs.fixed_image = target_file
        init.inputs.moving_image = source_file
        init.inputs.num_threads = nb_cpus
        init.inputs.out_file = init_file
        init.run()
        init_source_file = os.path.join(outdir, "init_source.nii.gz")
        at = ApplyTransforms()
        at.inputs.dimension = 3
        at.inputs.input_image = source_file
        at.inputs.reference_image = target_file
        at.inputs.output_image = init_source_file
        at.inputs.transforms = init_file
        at.inputs.interpolation = "BSpline"
        at.inputs.num_threads = nb_cpus
        at.run()
        # From https://github.com/ANTsX/ANTs/wiki/ITK-affine-transform-conversion
        # From https://github.com/netstim/leaddbs/blob/master/helpers/ea_antsmat2mat.m
        init_mat = loadmat(init_file)
        afftransform = init_mat["AffineTransform_double_3_3"]
        m_center = init_mat["fixed"]
        mat = numpy.eye(4)
        mat[:3, :3] = afftransform[:9].reshape(3, 3).T
        m_translation = afftransform[9:]
        m_offset = m_translation + m_center - numpy.dot(mat[:3, :3], m_center)
        mat[:3, 3] = m_offset.squeeze()
        mat = numpy.linalg.inv(mat)
        #rotation = swap_affine(axes="LPS")
        #mat = numpy.dot(rotation, mat)
        ras_to_lps = numpy.ones((4, 4))
        ras_to_lps[2, :2] = -1
        ras_to_lps[:2, 2:] = -1
        mat = mat * ras_to_lps
        print(mat)
        trans, rot, zoom, shear = decompose44(mat)
        euler = numpy.asarray(mat2euler(rot, axes='sxyz')) * 180. / numpy.pi
        print(trans, euler, zoom, shear)
        align_file = os.path.join(outdir, "align.com")
        im = nibabel.load(source_file)
        orig = im.affine[:3, 3]
        print(orig)
        shift = orig - trans
        with open(align_file, "wt") as of:
            of.write("set registration-translations {0} \n".format(
                " ".join([str(elem) for elem in shift])))

    # Change current working directory
    cwd = os.getcwd()
    os.chdir(outdir)

    # Only support uncompressed Nifti
    source_file = ungzip_file(source_file, prefix="", outdir=outdir)
    target_file = ungzip_file(target_file, prefix="", outdir=outdir)

    # Create jip environment
    jip_envriron = os.environ
    jip_envriron["JIP_HOME"] = os.path.dirname(jipdir)
    if "PATH" in jip_envriron:
        jip_envriron["PATH"] = jip_envriron["PATH"] + ":" + jipdir
    else:
        jip_envriron["PATH"] = jipdir

    # Copy source file
    align_file = os.path.join(outdir, "align.com")
    cmd = ["align", source_file, "-t", target_file]
    if auto:
        if non_linear:
            auto_cmd = cmd + ["-L", "111111111111", "-W", "111", "-a"]
        else:
            auto_cmd = cmd + ["-L", "111111000000", "-W", "000", "-A"]
        if os.path.isfile(align_file):
            auto_cmd += ["-I"]
        subprocess.call(auto_cmd, env=jip_envriron)  
    else:
        print(" ".join(cmd))
        subprocess.call(cmd, env=jip_envriron)
    if not os.path.isfile(align_file):
        raise ValueError(
            "No 'align.com' file in '{0}' folder. JIP has probably failed: "
            "'{1}'".format(outdir, " ".join(cmd)))

    # Get the apply nonlinear deformation jip batches
    aplly_nonlin_batch = os.path.join(
        os.path.dirname(preproc.__file__), "resources", "apply_nonlin.com")
    aplly_inv_nonlin_batch = os.path.join(
        os.path.dirname(preproc.__file__), "resources", "apply_inv_nonlin.com")

    # Resample the source image
    register_file = os.path.join(outdir, prefix + os.path.basename(source_file))
    cmd = ["jip", aplly_nonlin_batch, source_file, register_file]
    subprocess.call(cmd, env=jip_envriron)

    # Apply mask
    if os.path.isfile(register_file + ".gz"):
        os.remove(register_file + ".gz")
    register_mask_fileroot = os.path.join(
        outdir, "m" + prefix + os.path.basename(source_file).split(".")[0])
    register_mask_file = apply_mask(
        input_file=register_file,
        output_fileroot=register_mask_fileroot,
        mask_file=target_file,
        fslconfig=fslconfig)

    # Send back masked image to original space
    register_mask_file = ungzip_file(register_mask_file, prefix="",
                                     outdir=outdir)
    native_masked_file = os.path.join(
        outdir, "n" + prefix + os.path.basename(register_mask_file))
    cmd = ["jip", aplly_inv_nonlin_batch, register_mask_file,
           native_masked_file]
    subprocess.call(cmd, env=jip_envriron)              

    # Restore current working directory and gzip output
    os.chdir(cwd)
    register_file = gzip_file(
        register_file, prefix="", outdir=outdir,
        remove_original_file=True)
    register_mask_file = gzip_file(
        register_mask_file, prefix="", outdir=outdir,
        remove_original_file=True)
    native_masked_file = gzip_file(
        native_masked_file, prefix="", outdir=outdir,
        remove_original_file=True)

    return register_file, register_mask_file, native_masked_file, align_file


def apply_jip_align(apply_to_files, align_with, outdir, jipdir, prefix="w",
                    apply_inv=False):
    """ Apply a jip deformation.

    Parameters
    ----------
    apply_to_files: list of str (mandatory)
        a list of image path to apply the computed deformation.
    align_with: str ot list of str (mandatory)
        the alignement file containind the deformation parameters. Expect
        an 'align.com' file.
    outdir: str (mandatory)
        the destination folder.
    jipdir: str (mandatory)
        the jip binary path.
    prefix: str (optional, default 'w')
        prefix the generated file with this character.
    apply_inv: bool (optional, default False)
        if set apply the inverse deformation.

    Returns
    -------
    deformed_files: list of str
        the deformed input files.
    """
    # Check input parameters
    if not isinstance(align_with, list):
        align_with = [align_with]
    if not isinstance(apply_to_files, list):
        raise ValueError("The 'apply_to_files' function parameter must "
                         "contains a list of files.")
    for path in apply_to_files + align_with:
        if not os.path.isfile(path):
            raise ValueError("'{0}' is not a valid file.".format(path))

    # Get the apply nonlinear deformation jip batches
    if apply_inv:
        batch = os.path.join(os.path.dirname(preproc.__file__), "resources",
                             "apply_inv_nonlin.com")
    else:
        batch = os.path.join(os.path.dirname(preproc.__file__), "resources",
                             "apply_nonlin.com")
    batch = os.path.abspath(batch)

    # Create jip environment
    jip_envriron = os.environ
    jip_envriron["JIP_HOME"] = os.path.dirname(jipdir)
    if "PATH" in jip_envriron:
        jip_envriron["PATH"] = jip_envriron["PATH"] + ":" + jipdir
    else:
        jip_envriron["PATH"] = jipdir

    # Apply the jip deformation
    deformed_files = []
    cwd = os.getcwd()
    for path in apply_to_files:
        extra_file = ungzip_file(path, prefix="", outdir=outdir)
        deformed_file = os.path.join(
            outdir, prefix + os.path.basename(extra_file))
        for align_file in align_with:
            os.chdir(os.path.dirname(align_file))
            cmd = ["jip", batch, extra_file, deformed_file]
            subprocess.call(cmd, env=jip_envriron)
            extra_file = deformed_file
        deformed_file = gzip_file(
            deformed_file, prefix="", outdir=outdir, remove_original_file=True)
        deformed_files.append(deformed_file)
    os.chdir(cwd)

    return deformed_files


def check_jip_install(jipdir):
    """ Simple function to test if the JIP software bibaries are present.

    Parameters
    ----------
    jip_dir: str
        the JIP binaries directory.

    Returns
    -------
    status: bool
        the JIP installation status.
    """
    status = True
    for name in ("jip", "align"):
        binary_path = os.path.join(jipdir, name)
        if not os.path.isfile(binary_path):
            status = False
            break
    return status


def resample_image(source_file, target_file, out_file,
                   fslconfig=DEFAULT_FSL_PATH):
    """ Resample a source image to fit witha target image.

    Parameters
    ----------
    source_file: str
        the source Nifti image.
    target_file: str
        the target Nifti image.
    out_file: str
        the destination file.
    fslconfig: str (optional)
        the FSL .sh configuration file.
    """
    flirt(
        in_file=source_file,
        ref_file=target_file,
        init=None, cost="corratio",
        applyxfm=True,
        out=out_file,
        interp="nearestneighbour",
        shfile=fslconfig)
    return out_file
    



##########################################################################
# NSAp - Copyright (C) CEA, 2017
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################


"""
Tools to play with affine matrices.
"""


# Package import
from transforms3d.quaternions import mat2quat, quat2axangle

# Third party import
import numpy
import scipy.linalg

# Global parameters
MAX_ANGLE = 1e10 * 2 * numpy.pi
SMALL_ANGLE = 1e-30


def matrix44_to_vec12(affine):
    """ Decompose a 4x4 matrix describing a homogenous affine transform
    into a 12-sized vector composed of the translations, rotations, zooms,
    shears.

    Decomposing a matrix into simple transformations* by Spencer
    W. Thomas, pp 320-323 in *Graphics Gems II*, James Arvo (editor),
    Academic Press, 1991, ISBN: 0120644819.

    Parameters
    ----------
    affine: array (4, 4)
        the input affine matrix.

    Returns
    -------
    vec12: array (12, )
        the 12-sized vector of natural affine parameters.
    """
    # Initilize the output vector
    vec12 = numpy.zeros((12, ))

    # Extract translation
    vec12[0:3] = affine[:3, 3]

    # Extract x-scale and norm: for this, take the length of the first column
    # vectors e1
    RZS = affine[:-1, :-1]
    e1, e2, e3 = RZS.T
    sx = numpy.sqrt(numpy.sum(e1**2))
    e1 /= sx

    # Extract shear: orthogonalize e2 with respect to e1
    sx_sxy = numpy.dot(e1, e2)
    e2 -= sx_sxy * e1
    sxy = sx_sxy / sx

    # Extract y-scale and norm
    sy = numpy.sqrt(numpy.sum(e2**2))
    e2 /= sy

    # Extract shear: orthogonalize e3 with respect to e1 and e2
    sx_sxz = numpy.dot(e1, e3)
    sy_syz = numpy.dot(e2, e3)
    e3 -= (sx_sxz * e1 + sy_syz * e2)
    sxz = sx_sxz / sx
    syz = sy_syz / sy

    # Extract z-scale and norm
    sz = numpy.sqrt(numpy.sum(e3**2))
    e3 /= sz

    # Extract rotation
    rotation = numpy.array([e1, e2, e3]).T
    if numpy.linalg.det(rotation) < 0:
        sx *= -1
        rotation[:, 0] *= -1
    
    # Extract rotation angles
    angles = rotation_mat2vec(rotation)

    # Construct output vector
    vec12[3:6] = angles
    vec12[6:9] = numpy.array([sx, sy, sz])
    vec12[9:12] = numpy.array([sxy, sxz, syz])

    return vec12


def rotation_mat2vec(rotation):
    """ Rotation vector from rotation matrix.

    Parameters
    ----------
    rotation: array (3, 3)
        rotation matrix.

    Returns
    -------
    vec: array (3, )
        rotation vector, where norm of vec is the angle theta, and the
        axis of rotation is given by vec / theta.
    """
    ax, angle = quat2axangle(mat2quat(rotation))
    return ax * angle


def rotation_vec2mat(r):
    """ The rotation matrix is given by the Rodrigues formula:

    .. math::

        R = Id + sin(theta)*Sn + (1-cos(theta))*Sn^2

    with:

    .. math::

               0  -nz  ny
        Sn =   nz   0 -nx
              -ny  nx   0

    where n = r / ||r||

    In case the angle ||r|| is very small, the above formula may lead
    to numerical instabilities. We instead use a Taylor expansion
    around theta=0:

    .. math::

        R = I + sin(theta)/tetha Sr + (1-cos(theta))/teta2 Sr^2

    leading to:

    .. math::

        R = I + (1-theta2/6)*Sr + (1/2-theta2/24)*Sr^2

    To avoid numerical instabilities, an upper threshold is applied to
    the angle. It is chosen to be a multiple of 2*pi, hence the
    resulting rotation is then the identity matrix. This strategy warrants
    that the output matrix is a continuous function of the input vector.

    Parameters
    ----------
    vec: array (3, )
        rotation vector, where norm of vec is the angle theta, and the
        axis of rotation is given by vec / theta.

    Returns
    -------
    rotation: array (3, 3)
        rotation matrix.
    """
    theta = numpy.sqrt(numpy.sum(r ** 2))
    if theta > MAX_ANGLE:
        return numpy.eye(3)
    elif theta > SMALL_ANGLE:
        n = r / theta
        Sn = numpy.array([[0, -n[2], n[1]], [n[2], 0, -n[0]], [-n[1], n[0], 0]])
        R = numpy.eye(3) + numpy.sin(theta) * Sn\
            + (1 - numpy.cos(theta)) * numpy.dot(Sn, Sn)
    else:
        Sr = numpy.array([[0, -r[2], r[1]], [r[2], 0, -r[0]], [-r[1], r[0], 0]])
        theta2 = theta * theta
        R = numpy.eye(3) + (1 - theta2 / 6.) * Sr\
            + (.5 - theta2 / 24.) * numpy.dot(Sr, Sr)
    return R

